import unittest
import torch
from irisml.tasks.make_image_text_contrastive_model import Task


class TestMakeImageTextContrastiveModel(unittest.TestCase):
    def test_simple(self):
        image_model = torch.nn.Linear(3, 3)
        text_model = torch.nn.Linear(3, 3)

        inputs = Task.Inputs(image_model, text_model)
        outputs = Task(Task.Config(loss='clip')).execute(inputs)
        loss = outputs.model.training_step((torch.rand(2, 3), torch.rand(2, 3)), torch.tensor([0, 1]))
        self.assertEqual(len(loss['loss'].shape), 0)

        outputs = Task(Task.Config(loss='unicl')).execute(inputs)
        loss = outputs.model.training_step((torch.rand(2, 3), torch.rand(2, 3)), torch.tensor([0, 1]))
        self.assertEqual(len(loss['loss'].shape), 0)



import unittest
import torch
from irisml.tasks.get_targets_from_dataset import Task


class TestGetTargetsFromDataset(unittest.TestCase):
    def test_simple(self):
        class FakeDataset(torch.utils.data.Dataset):
            def __init__(self, data):
                super().__init__()
                self._data = data

            def __len__(self):
                return len(self._data)

            def __getitem__(self, index):
                return self._data[index]

        data = [('image0', 0), ('image1', 2), ('image2', 4)]
        inputs = Task.Inputs(FakeDataset(data))
        outputs = Task(Task.Config()).execute(inputs)

        targets = outputs.targets

        self.assertIsInstance(targets, torch.Tensor)
        self.assertEqual(targets.tolist(), [0, 2, 4])



import dataclasses
import itertools
import logging
import math
import multiprocessing
import queue
import typing
import torch.nn
import torch.utils.data
import irisml.core
from irisml.tasks.train.build_dataloader import build_dataloader
from irisml.tasks.train.ddp_utils import spawn_processes, recreate_dataloader_with_new_sampler

logger = logging.getLogger(__name__)


class Predictor:
    def __init__(self, model, device=torch.device('cpu')):
        self._model = model
        self._device = device
        self._predictor = getattr(model, 'predictor', None)

    @property
    def device(self):
        return self._device

    @torch.no_grad()
    def predict(self, dataloader):
        self._model.to(self.device)

        was_training = self._model.training
        self._model.eval()

        results = []
        targets = []
        for batch_index, batch in enumerate(dataloader):
            batch_results, batch_targets = self.prediction_step(batch, batch_index)
            results.append(batch_results)
            targets.append(batch_targets)

        if was_training:
            self._model.train()

        self._model.to(torch.device('cpu'))
        return self._move_to_cpu(self._aggregate_results(results)), self._move_to_cpu(self._aggregate_results(targets))

    def prediction_step(self, batch, batch_index):
        inputs, targets = batch
        inputs = self._to_device(inputs)
        if hasattr(self._model, 'prediction_step'):
            batch_results = self._model.prediction_step(inputs)
        elif self._predictor is not None:
            outputs = self._model(inputs)
            batch_results = self._predictor(outputs)
        else:
            batch_results = self._model(inputs)
        return batch_results, targets

    def _aggregate_results(self, results: typing.List):
        if torch.is_tensor(results[0]):
            if all(results[0].shape[1:] == r.shape[1:] for r in results):
                return torch.cat(results)
            else:
                return list(itertools.chain.from_iterable(list(x) for x in results))
        elif isinstance(results[0], list):
            return list(itertools.chain(*results))
        else:
            raise RuntimeError(f"Unexpected result types: {results[0]}")

    def _to_device(self, data):
        if isinstance(data, list):
            return [self._to_device(d) for d in data]
        elif isinstance(data, tuple):
            return tuple(self._to_device(d) for d in data)
        elif hasattr(data, 'to'):
            return data.to(self.device, non_blocking=True)
        return data

    @staticmethod
    def _move_to_cpu(value):
        if torch.is_tensor(value):
            return value.to(torch.device('cpu'))
        elif isinstance(value, list):
            return [Predictor._move_to_cpu(x) for x in value]
        elif isinstance(value, (int, float, str)):
            return value
        else:
            raise ValueError(f"Unexpected value type: {type(value)}")


class SubsetSequentialSampler(torch.utils.data.Sampler):
    def __init__(self, data_source, start_index, end_index):
        self._start_index = start_index
        self._end_index = end_index
        assert start_index < end_index
        super().__init__(data_source)

    def __len__(self):
        return self._end_index - self._start_index

    def __iter__(self):
        return iter(range(self._start_index, self._end_index))


class DistributedPredictor(Predictor):
    def __init__(self, model, num_processes, device=torch.device('cpu')):
        self._num_processes = num_processes
        super().__init__(model, device)

    def predict(self, dataloader):
        mp_queue = multiprocessing.get_context('spawn').JoinableQueue(self._num_processes)
        num_examples_per_process = math.ceil(len(dataloader.dataset) / self._num_processes)
        context = spawn_processes(self._predict_on_new_process, args=(dataloader, num_examples_per_process, mp_queue), nprocs=self._num_processes)
        all_results = {}

        while len(all_results) < self._num_processes:
            try:
                index, results = mp_queue.get(timeout=5)
                assert index not in all_results
                all_results[index] = results
            except queue.Empty:
                if not all(p.is_alive() for p in context.processes):
                    context.join(1)
                    raise RuntimeError("Failed to collect prediction results from the child processes.")

        aggregated_results = self._aggregate_results([all_results[i][0] for i in range(self._num_processes)])
        aggregated_targets = self._aggregate_results([all_results[i][1] for i in range(self._num_processes)])

        for _ in range(self._num_processes):
            mp_queue.task_done()

        context.join(5)

        return aggregated_results, aggregated_targets

    def _predict_on_new_process(self, process_index, dataloader, num_examples_per_process, mp_queue):
        if self._device.type == 'cuda':
            self._device = torch.device('cuda', process_index)
        start_index = num_examples_per_process * process_index
        end_index = min(len(dataloader.dataset), start_index + num_examples_per_process)
        logger.info(f"Launched a process {process_index} on device {self._device} for batches [{start_index}, {end_index}].")
        sampler = SubsetSequentialSampler(dataloader.dataset, start_index, end_index)
        subset_dataloader = recreate_dataloader_with_new_sampler(dataloader, sampler)
        try:
            results = super().predict(subset_dataloader)
        except Exception as e:
            logger.exception(f"Exception happend while predicting. {e}")
            raise
        logger.info(f"Completed predictions on the process {process_index}. Returning {len(results)} results.")

        mp_queue.put((process_index, results))
        mp_queue.join()


class Task(irisml.core.TaskBase):
    """Predict using a given model.

    This class assumes that the model.predictor returns a Tensor or a list of results.
    """
    VERSION = '0.1.1'

    @dataclasses.dataclass
    class Inputs:
        dataset: torch.utils.data.Dataset
        transform: torch.nn.Module
        model: torch.nn.Module

    @dataclasses.dataclass
    class Config:
        batch_size: int = 1
        device: typing.Optional[typing.Literal['cpu', 'cuda']] = None
        num_processes: int = 1

    @dataclasses.dataclass
    class Outputs:
        predictions: typing.Union[typing.List, torch.Tensor] = dataclasses.field(default_factory=list)
        targets: typing.Optional[typing.Union[typing.List, torch.Tensor]] = None

    def execute(self, inputs):
        results = self._predict(inputs)
        return self.Outputs(results[0], results[1])

    def _predict(self, inputs):
        dataloader = build_dataloader(inputs.dataset, inputs.transform, batch_size=self.config.batch_size, shuffle=False, drop_last=False)
        predictor = Predictor(inputs.model, self._get_device()) if self.config.num_processes == 1 else DistributedPredictor(inputs.model, self.config.num_processes, self._get_device())
        results = predictor.predict(dataloader)
        if len(inputs.dataset) != len(results[0]):
            raise RuntimeError(f"Couldn't get the expected number of prediction results. Expected: {len(inputs.dataset)}. Actual: {len(results[0])}")
        return results

    def _get_device(self) -> torch.device:
        """Get a torch device based on the configuration. If not specified explicitly, it uses cuda if available."""
        if self.config.device:
            device_name = self.config.device
        else:
            device_name = 'cuda' if torch.cuda.is_available() else 'cpu'
            logger.info(f"Training device is selected automatically: {device_name}. To specify the device manually, please set Config.device.")

        return torch.device(device_name)



class PluginBase:
    def on_train_start(self, trainer, model):
        pass

    def on_train_end(self, trainer, model):
        pass

    def on_train_epoch_start(self, trainer, model, epoch_index) -> bool:
        """If False is returned, the training will be stopped. Useful for early stopping."""
        return True

    def on_train_epoch_end(self, trainer, model, epoch_index):
        pass

    def on_train_batch_start(self, trainer, model, batch, batch_index):
        """Returns a modified batch."""
        return batch

    def on_train_batch_end(self, trainer, model, loss, batch, batch_index):
        pass

    def on_train_backward_start(self, trainer, model, loss):
        """Returns a modified loss."""
        return loss

    def on_train_backward_end(self, trainer, model):
        pass

    def on_optimizer_step_start(self, trainer, model, optimizer):
        """Returns an updated optimizer."""
        return optimizer

    def on_validation_start(self, trainer, model):
        pass

    def on_validation_end(self, trainer, model):
        pass

    def on_validation_batch_start(self, trainer, model, batch, batch_index):
        """Returns a modified batch."""
        return batch

    def on_validation_batch_end(self, trainer, model, loss, batch, batch_index):
        pass

    def forward_context(self):
        pass



import copy
import dataclasses
import typing
import irisml.core
import torch


class Task(irisml.core.TaskBase):
    """Append a classifier model to a given model. A predictor and a loss module will be added, too.

    Inputs:
        model (torch.nn.Module): An encoder model
        classifier (torch.nn.Module): A classifier model. The input shape of this module must match with the outptu shape of the encoder model.
    """
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Inputs:
        model: torch.nn.Module
        classifier: torch.nn.Module

    @dataclasses.dataclass
    class Config:
        task_type: typing.Literal['multiclass_classification', 'multilabel_classification']

    @dataclasses.dataclass
    class Outputs:
        model: torch.nn.Module = None

    def execute(self, inputs):
        model = ClassifierModel(copy.deepcopy(inputs.model), copy.deepcopy(inputs.classifier), self.config.task_type)
        return self.Outputs(model)


class ClassifierModel(torch.nn.Module):
    MODULES = {
        'multiclass_classification': (torch.nn.CrossEntropyLoss(), torch.nn.Softmax(1)),
        'multilabel_classification': (torch.nn.BCEWithLogitsLoss(), torch.nn.Sigmoid()),
    }

    def __init__(self, model, classifier, task_type):
        super().__init__()
        self._model = model
        self._classifier = classifier
        self._criterion, self._predictor = self.MODULES[task_type]

    def forward(self, *args):
        return self._classifier(self._model(*args))

    @property
    def criterion(self):
        return self._criterion

    @property
    def predictor(self):
        return self._predictor



import ast
import importlib
import logging
from irisml.core.variable import replace_variables

logger = logging.getLogger(__name__)


def load_plugin(plugin_expr, context):
    plugin_name, plugin_args = _parse_plugin_expr(plugin_expr, context)

    try:
        plugin_module = importlib.import_module('irisml.tasks.train.plugins.' + plugin_name)
    except ModuleNotFoundError as e:
        raise RuntimeError(f"Plugin {plugin_name} is not found.") from e

    plugin_class = getattr(plugin_module, 'Plugin', None)
    instance = plugin_class(*plugin_args)
    logger.debug(f"Loaded a plugin: {plugin_expr}")
    return instance


def _parse_plugin_expr(plugin_expr, context):
    parsed = ast.parse(plugin_expr)
    if len(parsed.body) != 1:
        raise ValueError(f"Plugin description cannot have multiple expressions: {plugin_expr}")

    if isinstance(parsed.body[0].value, ast.Name):
        plugin_name = parsed.body[0].value.id
        plugin_args = []
    elif isinstance(parsed.body[0].value, ast.Call):
        plugin_name = parsed.body[0].value.func.id
        ast_args = parsed.body[0].value.args
        if not all(isinstance(a, ast.Constant) for a in ast_args):
            raise ValueError(f"Only constant types can be used as plugin arguments: {ast.dump(ast_args)}")
        plugin_args = [a.value for a in ast_args]
        # Resolve variables in the plugin arguments.
        plugin_args = [context.resolve(replace_variables(a)) if isinstance(a, str) else a for a in plugin_args]
    else:
        raise ValueError(f"Unexpected plugin expression: {ast.dump(parsed)}")

    return plugin_name, plugin_args



import dataclasses
import typing
import irisml.core
import torch.distributed
import torch.nn
from irisml.tasks.train.ddp_utils import all_gather


class Task(irisml.core.TaskBase):
    """Make a model for image-text contrastive training.

    Currently this task supports two losses:
        clip (default): The loss function from https://arxiv.org/abs/2103.00020
        unicl: The loss function from https://arxiv.org/abs/2204.03610
    """
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Inputs:
        image_model: torch.nn.Module
        text_model: torch.nn.Module

    @dataclasses.dataclass
    class Config:
        loss: typing.Literal['clip', 'unicl'] = 'clip'
        image_feature_dim: typing.Optional[int] = None
        text_feature_dim: typing.Optional[int] = None
        projection_dim: typing.Optional[int] = 512

    @dataclasses.dataclass
    class Outputs:
        model: torch.nn.Module = None

    def execute(self, inputs):
        image_projection = torch.nn.Linear(self.config.image_feature_dim, self.config.projection_dim) if self.config.image_feature_dim and self.config.projection_dim else None
        text_projection = torch.nn.Linear(self.config.text_feature_dim, self.config.projection_dim) if self.config.text_feature_dim and self.config.projection_dim else None
        model = ImageTextContrastiveModel(inputs.image_model, inputs.text_model, image_projection, text_projection, self.config.loss)
        return self.Outputs(model)

    def dry_run(self, inputs):
        return self.execute(inputs)


class CLIPContrastiveCriterion(torch.nn.Module):
    def forward(self, features, targets):
        image_features, text_features, logit_scale = features
        logits = logit_scale * image_features @ text_features.t()
        image_loss = torch.nn.functional.cross_entropy(logits, torch.arange(len(logits), device=logits.device))
        text_loss = torch.nn.functional.cross_entropy(logits.t(), torch.arange(len(logits), device=logits.device))
        return (image_loss + text_loss) / 2.0


class UnifiedContrastiveCriterion(torch.nn.Module):
    def forward(self, features, targets):
        """
        Args: (image_features, text_features, logit_scal), targets
            image_features (torch.Tensor): Shape (N, feature_size)
            text_features (torch.Tensor): Shape (N, feature_size)
            logit_scale (torch.Tensor): A scalar
            targets (torch.Tensor): Shape (N, 1).
        """
        image_features, text_features, logit_scale = features
        logits = logit_scale * image_features @ text_features.t()
        targets = (targets.view(-1, 1) == targets.view(1, -1)).float()
        targets_sum = torch.sum(targets, dim=-1)  # For each sample, there is at least one positive targets.
        image_loss = torch.nn.functional.cross_entropy(logits, targets, reduction='none')
        text_loss = torch.nn.functional.cross_entropy(logits.t(), targets, reduction='none')
        return ((image_loss + text_loss) / targets_sum).mean() / 2.0


class ImageTextContrastiveModel(torch.nn.Module):
    def __init__(self, image_model, text_model, image_projection, text_projection, loss_name, logit_scale=2.659260036932778):
        """
        Args:
            image_model (torch.nn.Module): input: Tensor[N, C, H, W], output: Tensor[N, F]
            text_model (torch.nn.Module): input: Tensor[N, T], output: Tensor[N, F2]
            image_projection (torch.nn.Module): Applied to image_model's output. Optional.
            text_projection (torch.nn.Module): Applied to text_model's output. Optional.
            loss_name: clip or unicl
        Notes:
            math.log(1 / 0.07) = 2.659260036932778
        """
        super().__init__()

        # The task 'split_image_text_model' depends on these attributes.
        self.image_model = image_model
        self.text_model = text_model
        self.image_projection = image_projection or torch.nn.Identity()
        self.text_projection = text_projection or torch.nn.Identity()
        self.logit_scale = torch.nn.Parameter(torch.tensor(logit_scale))

        if loss_name == 'clip':
            self._criterion = CLIPContrastiveCriterion()
        elif loss_name == 'unicl':
            self._criterion = UnifiedContrastiveCriterion()
        else:
            raise RuntimeError

    @property
    def criterion(self):
        return self._criterion

    def forward(self, inputs):
        image_features = self.image_projection(self.image_model(inputs[0]))
        text_features = self.text_projection(self.text_model(inputs[1]))

        image_features = image_features / image_features.norm(p=2, dim=-1, keepdim=True)
        text_features = text_features / text_features.norm(p=2, dim=-1, keepdim=True)

        return image_features, text_features, self.logit_scale.exp()

    def training_step(self, inputs, targets):
        image_features, text_features, logit_scale = self(inputs)
        if torch.distributed.is_initialized():
            image_features = all_gather(image_features)
            text_features = all_gather(text_features)

        assert len(image_features.shape) == 2 and image_features.shape == text_features.shape, f"Unexpected feature shapes: {image_features.shape} {text_features.shape}"

        loss = self._criterion((image_features, text_features, logit_scale), targets)
        return {'loss': loss}

    def prediction_step(self, inputs):
        raise RuntimeError("Prediction is not supported. If you would like to try zero-shot classificaiton, see the task 'build_zero_shot_classifier'.")



import unittest
import torch
from irisml.core import Context
from irisml.tasks.export_onnx import Task


class TestExportOnnx(unittest.TestCase):
    def test_model_with_prediction_step(self):
        class FakeModel(torch.nn.Module):
            def __init__(self):
                super().__init__()
                self._model = torch.nn.Conv2d(3, 3, 3)
                self._predictor = torch.nn.Softmax(1)

            def prediction_step(self, x):
                return self._predictor(self(x))

            def forward(self, x):
                return self._model(x)

        model = FakeModel()

        task = Task(Task.Config(), Context())
        outputs = task.execute(Task.Inputs(model))
        self.assertIsNotNone(outputs)

    def test_model_with_predictor(self):
        class FakeModel(torch.nn.Module):
            def __init__(self):
                super().__init__()
                self._model = torch.nn.Conv2d(3, 3, 3)
                self._predictor = torch.nn.Softmax(1)

            @property
            def predictor(self):
                return self._predictor

            def forward(self, x):
                return self._model(x)

        model = FakeModel()

        task = Task(Task.Config(), Context())
        outputs = task.execute(Task.Inputs(model))
        self.assertIsNotNone(outputs)

    def test_model_without_predictor(self):
        class FakeModel(torch.nn.Module):
            def __init__(self):
                super().__init__()
                self._model = torch.nn.Conv2d(3, 3, 3)

            def forward(self, x):
                return self._model(x)

        model = FakeModel()

        task = Task(Task.Config(), Context())
        outputs = task.execute(Task.Inputs(model))
        self.assertIsNotNone(outputs)



import logging
import multiprocessing
import os
import queue
import torch
from .ddp_utils import create_distributed_data_parallel_model, spawn_and_wait, recreate_dataloader_with_distributed_sampler
from .trainer import Trainer
from .plugins.ddp import Plugin as DDPPlugin

logger = logging.getLogger(__name__)


class DDPTrainer(Trainer):
    """Train with DistributedDataParallel.

    This class assumes single-node multi-device environment.
    """
    def __init__(self, model, lr_scheduler_factory, optimizer_factory, plugins=None, device=torch.device('cpu'), val_check_interval=None, num_processes=2):
        plugins = plugins or []
        plugins.append(DDPPlugin())
        super().__init__(model, lr_scheduler_factory, optimizer_factory, plugins, device, val_check_interval)
        if device.type == 'cuda' and torch.cuda.is_available() and num_processes > torch.cuda.device_count():
            raise RuntimeError(f"Distributed training is used with {num_processes} processes, but there are only {torch.cuda.device_count()} devices.")
        if num_processes == 1:
            raise RuntimeError("At least 2 processes are required for DDP training.")

        if device.type == 'cuda':
            # Crashes the process on errors. This setting has very little performance overhead.
            os.environ['NCCL_ASYNC_ERROR_HANDLING'] = '1'

        logger.info(f"Distributed training with {num_processes} processes.")
        self._num_processes = num_processes

    def train(self, train_dataloader, val_dataloader, num_epochs):
        """Spawn processes and run training on each process.

        To send tensors from a child process, the process must be kept alive until the tensors are loaded on the main process.
        """
        mp_queue = multiprocessing.get_context('spawn').JoinableQueue(1)

        if self.device.type == 'cuda':
            torch.cuda.empty_cache()  # Needs to remove the cache on this process since it is not usable from the new processes.

        context = spawn_and_wait(self._train_on_new_process, args=(train_dataloader, val_dataloader, num_epochs, mp_queue), nprocs=self._num_processes)

        logger.info("At least one of the child processes has completed. Now waiting for the outputs.")

        try:
            state_dict = mp_queue.get(timeout=60)  # If it couldn't get the state_dict in 60 seconds, something wrong must have happend.
            try:
                logger.info(list(state_dict.keys()))
                self.model.load_state_dict(state_dict)
            finally:
                mp_queue.task_done()  # Terminate the child process after loading the state_dict.
        except queue.Empty:
            context.join(1)  # If there was an exception information, let the library handle it.
            raise RuntimeError("Failed to get a trained model from the child process.")

        context.join()

    def _train_on_new_process(self, process_index, train_dataloader, val_dataloader, num_epochs, mp_queue):
        """Method that runs on a new process.

        This method returns immediately after the training if process_index != 0. If process_index == 0, it will wait for the main process to read the trained state_dict.
        """
        device_id = [process_index] if self._device.type == 'cuda' else None
        if self._device.type == 'cuda':
            self._device = torch.device('cuda', process_index)
            torch.cuda.set_device(process_index)

        backend = 'gloo' if device_id is None else 'nccl'
        logger.info(f"New process started on device {self._device}. backend={backend}, rank={process_index}")
        torch.distributed.init_process_group(backend, rank=process_index, world_size=self._num_processes)
        logger.info(f"Initialized the process group on rank {process_index}.")

        # The model needs to be on the target device. Otherwise, DistributedDataParallel() will stuck.
        self._model.to(self._device)

        self._model = create_distributed_data_parallel_model(self.model, device_id)

        distributed_train_dataloader = recreate_dataloader_with_distributed_sampler(train_dataloader)
        distributed_val_dataloader = val_dataloader and recreate_dataloader_with_distributed_sampler(val_dataloader)

        super().train(distributed_train_dataloader, distributed_val_dataloader, num_epochs)

        if process_index == 0:
            # All processes have the same model at this point. Sending it to the main process from the 0th process.
            state_dict = self._model.module.to(torch.device('cpu')).state_dict()
            logger.debug("Sending the state_dict from the process 0")
            mp_queue.put(state_dict)
            mp_queue.join()  # Wait until the state_dict is loaded by the main process.

        logger.info(f"Process {process_index} is completed.")



"""Early stopping for training.

To use this plugin, validation steps must be included in the training.
"""
import logging
from .plugin_base import PluginBase

logger = logging.getLogger(__name__)


class Plugin(PluginBase):
    def __init__(self, patience=2, min_delta=0.0):
        """
        Args:
            patience (int): Number of validation epochs with no improvement after which training will be stopped.
            min_delta (float): Minimum change in the loss to qualify as improvement
        """
        self._patience = patience
        self._min_delta = min_delta
        logger.info(f"Early stopping is enabled with {patience=}, {min_delta=}")

    def on_train_start(self, trainer, model):
        self._counter = 0
        self._early_stop = False
        self._best_loss = None

    def on_train_end(self, trainer, model):
        if self._best_loss is None:
            logger.warning("Early stopping was enabled but validation step was never called. Please enable validation to use early stopping.")

    def on_train_epoch_start(self, trainer, model, epoch_index):
        return not self._early_stop

    def on_validation_start(self, trainer, model):
        self._validation_sum_loss = 0
        self._validation_num_samples = 0

    def on_validation_batch_end(self, trainer, model, loss, batch, batch_index):
        self._validation_sum_loss += float(loss.item()) * len(batch[0])
        self._validation_num_samples += len(batch[0])

    def on_validation_end(self, trainer, model):
        if self._validation_num_samples == 0:
            logger.warning("Validation set doesn't have any samples. Please double check your dataset settings.")
            return

        mean_validation_loss = self._validation_sum_loss / self._validation_num_samples
        if self._best_loss is None or mean_validation_loss < self._best_loss - self._min_delta:
            self._best_loss = mean_validation_loss
            self._counter = 0
        else:
            self._counter += 1

        if self._counter > self._patience:
            logger.info(f"The validation loss has not improved for {self._counter} times. Stopping the training.")
            self._early_stop = True



import unittest
import torch
from irisml.tasks.train.build_dataloader import collate, Dataset


class TestDataLoader(unittest.TestCase):
    def test_collate(self):
        # Image only
        result = collate([torch.zeros((3, 64, 64), dtype=float) for _ in range(8)])
        self.assertIsInstance(result[0], torch.Tensor)
        self.assertIsNone(result[1])
        self.assertEqual(result[0].shape, (8, 3, 64, 64))

        result = collate([(torch.zeros((3, 64, 64), dtype=float), i) for i in range(8)])
        self.assertIsInstance(result[0], torch.Tensor)
        self.assertEqual(result[0].shape, (8, 3, 64, 64))
        self.assertIsInstance(result[1], torch.Tensor)
        self.assertEqual(result[1].shape, (8, ))

        result = collate([(torch.zeros((3, 64, 64), dtype=float), torch.tensor(i)) for i in range(8)])
        self.assertIsInstance(result[0], torch.Tensor)
        self.assertEqual(result[0].shape, (8, 3, 64, 64))
        self.assertIsInstance(result[1], torch.Tensor)
        self.assertEqual(result[1].shape, (8, ))

        # multilabel classification or object detection with no labels
        result = collate([(torch.zeros((3, 64, 64), dtype=float), []) for _ in range(8)])
        self.assertIsInstance(result[0], torch.Tensor)
        self.assertEqual(result[0].shape, (8, 3, 64, 64))
        self.assertIsInstance(result[1], list)
        self.assertFalse(any(result[1]))

        result = collate([(torch.zeros((3, 64, 64), dtype=float), [i]) for i in range(8)])
        self.assertIsInstance(result[0], torch.Tensor)
        self.assertEqual(result[0].shape, (8, 3, 64, 64))
        self.assertIsInstance(result[1], torch.Tensor)
        self.assertEqual(result[1].shape, (8, ))

        result = collate([(torch.zeros((3, 64, 64), dtype=float), [j for j in range(i)]) for i in range(8)])
        self.assertIsInstance(result[0], torch.Tensor)
        self.assertEqual(result[0].shape, (8, 3, 64, 64))
        self.assertIsInstance(result[1], list)
        self.assertTrue(torch.equal(result[1][0], torch.tensor([])))
        self.assertTrue(torch.equal(result[1][1], torch.tensor([0])))
        self.assertTrue(torch.equal(result[1][2], torch.tensor([0, 1])))

        # object detection
        result = collate([(torch.zeros((3, 64, 64), dtype=float), [[0, 0.1, 0.1, 0.2, 0.2] for _ in range(i)]) for i in range(8)])
        self.assertIsInstance(result[0], torch.Tensor)
        self.assertEqual(result[0].shape, (8, 3, 64, 64))
        self.assertIsInstance(result[1], list)
        self.assertIsInstance(result[1][0], torch.Tensor)
        self.assertIsInstance(result[1][1], torch.Tensor)
        self.assertEqual(result[1][0].shape, (0, 5))
        self.assertTrue(torch.equal(result[1][1], torch.tensor([[0, 0.1, 0.1, 0.2, 0.2]])))
        self.assertEqual(result[1][2].shape, (2, 5))

        result = collate([(torch.zeros((3, 64, 64), dtype=float), torch.tensor([[0, 0.1, 0.1, 0.2, 0.2] for _ in range(i)])) for i in range(8)])
        self.assertIsInstance(result[0], torch.Tensor)
        self.assertEqual(result[0].shape, (8, 3, 64, 64))
        self.assertIsInstance(result[1], list)
        self.assertIsInstance(result[1][0], torch.Tensor)
        self.assertIsInstance(result[1][1], torch.Tensor)
        self.assertEqual(result[1][1].shape, (1, 5))
        self.assertEqual(result[1][2].shape, (2, 5))

        # image-text
        result = collate([((torch.zeros((3, 64, 64), dtype=float), torch.zeros(16, dtype=int)), [3]) for _ in range(8)])
        self.assertIsInstance(result[0][0], torch.Tensor)  # image
        self.assertIsInstance(result[0][1], torch.Tensor)  # text
        self.assertIsInstance(result[1], torch.Tensor)  # targets

        # regression
        result = collate([(torch.zeros((3, 64, 64), dtype=float), [3.5]) for _ in range(8)])
        self.assertIsInstance(result[0], torch.Tensor)  # image
        self.assertIsInstance(result[1], torch.Tensor)  # targets

    def test_get_transform_args(self):
        # Make sure it doesn't throw an exception.
        Dataset([], self._fake_transform_method1)
        Dataset([], self._fake_transform_method2)

        def fake_transform_function1(x):
            return x

        def fake_transform_function2(x, y):
            return x, y

        Dataset([], fake_transform_function1)
        Dataset([], fake_transform_function2)

    def _fake_transform_method1(self, x):
        return x

    def _fake_transform_method2(self, x, y):
        return x, y



import dataclasses
import typing
import torch
import irisml.core


class Task(irisml.core.TaskBase):
    """Extract only targets from a given Dataset.

    If the targets are integers or tensors with same shape, a tensor will be returned. Otherwise, returns a list of targets.
    """
    VERSION = '0.1.0'

    @dataclasses.dataclass
    class Inputs:
        dataset: torch.utils.data.Dataset

    @dataclasses.dataclass
    class Outputs:
        targets: typing.Union[typing.List, torch.Tensor] = dataclasses.field(default_factory=list)

    def execute(self, inputs):
        targets = [t for _, t in inputs.dataset]
        if torch.is_tensor(targets[0]):
            shape = targets[0].shape
            if all(t.shape == shape for t in targets):
                targets = torch.stack(targets)
        elif isinstance(targets[0], int):
            targets = torch.tensor(targets)

        return self.Outputs(targets)



import unittest
import torch
from irisml.tasks.make_oversampled_dataset import Task


class TestMakeOversampledDataset(unittest.TestCase):
    def test_get_class_id(self):
        def assert_result(input_targets, expected_class_ids):
            outputs = Task._get_class_id((None, input_targets))
            self.assertEqual(set(outputs), set(expected_class_ids))

        assert_result(1, [1])
        assert_result(0, [0])
        assert_result([1, 2, 3], [1, 2, 3])
        assert_result(0.1, [])
        assert_result([0.1], [])
        assert_result([], [])
        assert_result([[]], [])
        assert_result([[3, 0.1, 0.1, 0.1, 0.1]], [3])
        assert_result(torch.tensor(1), [1])
        assert_result(torch.tensor([1]), [1])
        assert_result(torch.tensor([1, 2, 3]), [1, 2, 3])
        assert_result(torch.tensor([[3, 1.0, 2.0, 3.0, 4.0]]), [3])
        assert_result(torch.tensor(2.0), [])

    def test_calculate_weights(self):
        dataset = [(None, [1]), (None, [1]), (None, [0])]
        results = Task._calculcate_weights(dataset, 100, 0)
        self.assertEqual(results[0], results[1])
        self.assertGreater(results[2], results[0])

    def test_min_num_samples(self):
        dataset = [(None, [1]), (None, [1]), (None, [0])]
        outputs = Task(Task.Config(10)).execute(Task.Inputs(dataset))
        self.assertEqual(len(outputs.dataset), 10)
        self.assertIsNotNone(outputs.dataset[0])
        self.assertIsNotNone(outputs.dataset[9])

        dataset = [(None, [1]), (None, [1]), (None, [0])]
        outputs = Task(Task.Config(3)).execute(Task.Inputs(dataset))
        self.assertEqual(outputs.dataset, dataset)  # no oversampling

    def test_oversampling_rate(self):
        dataset = [(None, [1]), (None, [1]), (None, [0])]
        outputs = Task(Task.Config(0, 3)).execute(Task.Inputs(dataset))
        self.assertEqual(len(outputs.dataset), 9)

        dataset = [(None, [1]), (None, [1]), (None, [0])]
        outputs = Task(Task.Config(0, 1)).execute(Task.Inputs(dataset))
        self.assertEqual(outputs.dataset, dataset)  # no oversampling

    def test_balancing(self):
        dataset = [(None, [1]), (None, [1]), (None, [0])]
        outputs = Task(Task.Config(0, 3, balance=True)).execute(Task.Inputs(dataset))
        self.assertEqual(len(outputs.dataset), 9)



import unittest
import torch
from irisml.tasks.evaluate_accuracy import Task


class TestEvaluateAccuracy(unittest.TestCase):
    def test_simple(self):
        predictions = torch.Tensor([[0.1, 0.2, 0.1], [0.9, 0.2, 0.2], [0.4, 0.4, 0.8]])
        targets = torch.Tensor([1, 0, 2])
        outputs = Task(Task.Config()).execute(Task.Inputs(predictions, targets))

        self.assertEqual(outputs.accuracy, 1.0)

        targets = torch.Tensor([2, 1, 1])
        outputs = Task(Task.Config()).execute(Task.Inputs(predictions, targets))
        self.assertEqual(outputs.accuracy, 0.0)

    def test_list_targets(self):
        predictions = torch.Tensor([[0.1, 0.2, 0.1], [0.9, 0.2, 0.2], [0.4, 0.4, 0.8]])
        targets = torch.Tensor([[1], [0], [2]])
        outputs = Task(Task.Config()).execute(Task.Inputs(predictions, targets))

        self.assertEqual(outputs.accuracy, 1.0)



import unittest
import torch.nn
from irisml.tasks.split_image_text_model import Task


class TestSplitImageTextModel(unittest.TestCase):
    def test_simple(self):
        class FakeModel(torch.nn.Module):
            def __init__(self):
                super().__init__()
                self.image_model = torch.nn.Conv2d(3, 3, 3)
                self.text_model = torch.nn.Conv2d(3, 3, 3)
                self.image_projection = torch.nn.Identity()
                self.text_projection = torch.nn.Identity()

        inputs = Task.Inputs(FakeModel())
        outputs = Task(Task.Config()).execute(inputs)
        self.assertIsNotNone(outputs.image_model)
        self.assertIsNotNone(outputs.text_model)



import copy
import unittest
import torch
from irisml.tasks.train.ddp_trainer import DDPTrainer


class FakeDataset(torch.utils.data.Dataset):
    def __init__(self, data):
        self._data = data

    def __getitem__(self, index):
        return self._data[index]

    def __len__(self):
        return len(self._data)


class FakeModel(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.model = torch.nn.Conv2d(3, 3, 3)

    def forward(self, x):
        return torch.flatten(torch.nn.AdaptiveAvgPool2d(1)(self.model(x)), start_dim=1)

    @property
    def criterion(self):
        return torch.nn.CrossEntropyLoss()


def fake_optimizer_factory(model):
    return torch.optim.SGD(model.parameters(), lr=0.01)


def fake_lr_scheduler_factory(optimizer):
    return torch.optim.lr_scheduler.CosineAnnealingWarmRestarts(optimizer, 1)


class TestDDPTrainer(unittest.TestCase):
    def test_train_without_dataload_thread(self):
        dataset = FakeDataset([[torch.rand(3, 256, 256), torch.tensor(1)], [torch.rand(3, 256, 256), torch.tensor(2)]])
        train_dataloader = torch.utils.data.DataLoader(dataset, num_workers=0)
        model = FakeModel()
        original_state_dict = copy.deepcopy(model.state_dict())
        trainer = DDPTrainer(model, fake_lr_scheduler_factory, fake_optimizer_factory, num_processes=4)
        trainer.train(train_dataloader, None, 1)

        # The training parameter is updated on this process.
        for key in original_state_dict:
            self.assertFalse(torch.equal(original_state_dict[key], trainer.model.state_dict()[key]))
